/**
 * @file Ce fichier contient la classe de l'API du LDAP qui gère les opérations sur les groupes. Elle est destinée à être exportée pour être utilisée par les resolvers.
 * @author hawkspar
 */

import { ldapConfig, groupData, categories } from '../internal/config';
import {Basics} from '../internal/basics';
import {Tools} from '../internal/tools';

//------------------------------------------------------------------------------------------------------------------------
// Classes à exporter TBT
//------------------------------------------------------------------------------------------------------------------------

export class Group {
    /**
     * @memberof LDAP
     * @class Group
     * @classdesc Cette classe est une des deux classes exportables permettant de faire des opérations sur les groupes.
     * @summary Constructeur vide.
     */
    constructor() {}
     
    /**
     * @memberof LDAP
     * @summary Fonction qui renvoit toutes les infos relatives à un groupe particulier.
     * @desc Cette fonction utilise {@link Tools.peek} avec l'interface {@link groupData}.
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(groupData)} Informations recueillies ; renvoie une liste de dictionnaire avec le profil complet du groupe au format {@link groupData}.
     * @static
     * @async
     */
    static async peek(gid: string) : Promise<groupData> {
        try {
            return Tools.peek<groupData>("group", gid, groupData);
        }
        catch(err) {
            throw "Erreur lors d'une recherche d'informations sur un groupe.";
        }
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui retrouve le groupe qui ressemblent à l'input et qui correspond au type fourni. Etape 0 vers un vrai TOL (Trombino On Line) pour les groupes.
     * @desc Cette fonction utilise {@link Tools.search}.
     * @arg {groupData} input - String entré par l'utilisateur qui ressemble au nom du groupe.
     * @return {Promise(string[])} Liste des gid dont le nom ressemble à l'input.
     * @static
     * @async
    */
    static async search(data: groupData) : Promise<string[]> {
        try {
            return Tools.search("group", data);
        }
        catch(err) {
            throw "Erreur lors de la recherche approximative d'un groupe.";
        }
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de rajouter un administrateur à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}. Le nouvel administrateur devient aussi membre et porte-parole du groupe,
     * mais hérite aussi de son statut d'administrateur sur tous les groupes qui héritent du sien.
     * @arg {string} uid - Identifiant du membre futur admin
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addAdmin(uid: string, gid: string): Promise<boolean> {
        return  await Tools.add(uid, gid, "members") &&
                await Tools.add(uid, gid, "speakers") &&
                await Tools.add(uid, gid, "admins");
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de supprimer un administrateur.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.remove}.
     * Elle ne remonte pas les échelons, car cela permettrait à un admin d'un petit groupe de supprimer un admin d'un grand. 
     * @arg {string} uid - Identifiant de l'admin à dégrader, supposé membre
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async remAdmin(uid: string, gid: string): Promise<boolean> {
        return Tools.remove(uid, gid, "admins");
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de rajouter un porte-parole à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}. Elle rajoute également l'utilisateur au groupe.
     * @arg {string} uid - Identifiant du membre futur porte-parole
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addSpeaker(uid: string, gid: string): Promise<boolean> {
        return  await Tools.add(uid, gid, "members") &&
                await Tools.add(uid, gid, "speakers");
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de rétrograder un membre du stade de porte-parole d'un groupe au stade d'utilisateur.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.remove}. Elle dégrade aussi d'un éventuel statut d'administrateur.
     * @arg {string} uid - Identifiant de l'admin à dégrader (pas supposé valide)
     * @arg {string} gid - Identifiant du groupe
     * @return {boolean} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async remSpeaker(uid: string, gid: string): Promise<boolean> {
        return  await Tools.remove(uid, gid, "admins") &&
                await Tools.remove(uid, gid, "speakers");
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet d'ajouter un utilisateur à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}.
     * @arg {string} uid - Identifiant de l'utilisateur à ajouter
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addMember(uid: string, gid: string) : Promise<boolean> {
        return Tools.add(uid, gid, "members");
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de supprimer un membre existant d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}.
     * Cette fonction supprime tous les droits de l'utilisateur sur le groupe, mais aussi sur les groupes sources si sont statut de membre était hérité.
     * @arg {string} uid - Identifiant de l'ex-membre (pas supposé valide)
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async remMember(uid: string, gid: string): Promise<boolean> {
        let stack = [];
        let res = true;
        let visited = {};
        stack.push(gid);
        while (stack.length>0) {
            let cur_id = stack.pop();
            if (visited[cur_id] == undefined) {
                visited[cur_id] = true;
                res = res   &&  await Tools.remove(uid, cur_id, "admins") &&
                                await Tools.remove(uid, cur_id, "speakers") &&
                                await Tools.remove(uid, cur_id, "members");
                stack.concat(await Basics.searchSingle("group", ldapConfig.group.childs, cur_id));
            }
        }
        return  res;
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet d'ajouter un sympathisant à un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}.
     * @arg {string} uid - Identifiant de l'utilisateur à ajouter
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async addFollower(uid: string, gid: string) : Promise<boolean> { return Tools.add(uid, gid, "followers"); }

    /**
     * @memberof LDAP
     * @summary Fonction qui permet de supprimer un sympathisant d'un groupe.
     * @desc Cette fonction fait essentiellement appel à {@link Tools.add}.
     * @arg {string} uid - Identifiant de l'ex-sympathisant (pas supposé valide)
     * @arg {string} gid - Identifiant du groupe
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async remFollower(uid: string, gid: string): Promise<boolean> { return Tools.remove(uid, gid, "followers"); }

    /**
     * @memberof LDAP
     * @summary Fonction qui créé un nouveau groupe dans le LDAP.
     * @desc Cette fonction fait une utilisation massive d'eval pour anonymiser son code ; c'est mal et cela suppose que beaucoup de soins ont été pris lors de
     * l'escape de ses paramètres. Appelle {@link LDAP.add} et {@link LDAP.change}, mais aussi {@link Tools.add}
     * pour gérer les groupes du nouvel utilisateur. Attention une manip FOIREUSE est cachée dedans.
     * @arg {groupData} data - Dictionnaire des informations utilisateurs (voir détail des champs dans ldapConfig.json)
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async create(data: groupData) : Promise<boolean> {
        // Calcul d'un dictionnaire d'ajout
        let vals = {};

        // gid de base généré à partir du nom standardisé, pas à partir de l'entrée 'gid' !
        try {
            Tools.generateReadableId(data['name']).then(id => {
                vals[ldapConfig.key_id]=id;
                vals[ldapConfig.group['name']]=id;
            });
        }
        catch(err) {
            throw "Erreur lors de la génération d'un hruid pour créer un nouveau groupe.";
        }

        let gid: string = vals[ldapConfig.key_id];

        // Ecriture de toutes les valeurs directement inscrites dans le LDAP
        for (let key_att in data) { vals[ldapConfig.group[key_att]]=data[key_att] };

        // Appel à la fonction de base
        if (!await Basics.add("group", vals)) {
            throw "Erreur lors de la création d'une nouvelle feuille dans l'arbre des groupes.";
        }
        // Certains champs nécessitent de petits calculs
        let vals2={};

        // ?!
        vals2[ldapConfig.group['password']] = "{CRYPT}"+data['password'];

        // Génération id aléatoire et test contre le LDAP
        try {
            Tools.generateId(ldapConfig.group["idNumber"], "group").then(id => { vals2[ldapConfig.group['idNumber']]=id; });
        }
        catch(err) {
            throw "Erreur lors de la génération d'un id numérique pour créer un nouveau groupe.";
        }

        // Code root
        vals2[ldapConfig.group['cleanFullName']]=data['name'].replace(':', ';').toLowerCase().normalize('UFD');

        // Inscription des valeurs calculées par effet de bord
        if (!await Basics.change("group", gid, "add", vals2)) {
            throw "Erreur lors de l'ajout des valeurs intelligentes du nouveau groupe.";
        }

        ["posixGroup", "brGroup"].forEach(cst => {
            let vals3={};
            vals3[ldapConfig.group['classes']]=cst;
            Basics.change("group", gid, "add", vals3).then(res => {
                if (!res) { throw "Erreur lors de l'ajout des valeurs constantes du nouveau groupe."; }
            });
        });

        // Utilisation des fonctions adaptées pour assurer la cohérence de l'ensemble
        for (let cat in categories) {
            data[cat].forEach(uid => {
                Tools.add(uid, gid, cat).then(res => {
                    if (!res) { throw "Erreur de l'ajout d'un membre au nouveau groupe."; }
                });
            });
        }
        return true;
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui supprime un groupe du LDAP.
     * @desc Cette fonction commence par gérer les groupes du membre puis le supprime entièrement. A modifier une fois que le LDAP incluerait les groupes administres par une utilisateur.
     * Appelle {@link LDAP.clear} bien sûr, mais aussi {@link Tools.remove} pour gérer les groupes de l'utilisateur sortant.
     * @arg {string} gid - gid de la victime
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async delete(gid: string): Promise<boolean> {
        try {
            // Gestion des catégories en bloc d'abord
            let profil = await Group.peek(gid);
            for (let cat of ["admins", "speakers", "members", "followers"]) {
                profil[ldapConfig.group[cat]].forEach(async function quickPartRemUser(uid: string) {
                    // Modification des profils de tous les utilisateurs
                    let lg = await Tools.get(uid, "user", cat);
                    // Vérifie que l'utilisateur est pas déjà viré pour users
                    if (lg.includes(gid)) {
                        // Supprime tous les groupes
                        if (!await Basics.change("user", uid, "del", ldapConfig.user[cat])) {
                            throw "Erreur lors de la suppression de tous les groupes du membre.";
                        }
                        // Les rajoute un par un, sauf pour le groupe supprimé
                        lg.forEach(id => {
                            if (id!=gid) {
                                Tools.add(uid, id, cat).then(res => {
                                    if (!res) { throw "Erreur lors du ré-ajout des autres groupes"; }
                                });
                            }
                        });
                    }
                });
            }
            // Elimination
            if (!await Basics.clear("group",gid)) { throw "Erreur lors de la suppression de la feuille dans l'arbre des groupes."; }
            return true;
        }
        catch(err) {
            throw "Erreur lors de l'obtention du profil d'un groupe pour le supprimer.";
        }
    }

    /**
     * @memberof LDAP
     * @summary Fonction qui édite un groupe existant dans le LDAP. Sans influence sur ses membres ou admins. 
     * @desc Appelle {@link Tools.edit}.
     * @arg {groupData} data - Dictionnaire des informations du groupe.
     * @return {Promise(boolean)} `true` si la modification s'est bien déroulée, false sinon
     * @async
     * @static
     */
    static async edit(data: groupData) : Promise<boolean> {
        try {
            return Tools.edit("group",data);
        }
        catch(err) {
            throw "Erreur lors de la modification d'un groupe.";
        }
    }
}